# Copyright (C) 2008 AG Projects. See LICENSE for details.
#

"""Implementation of a OpenSIPS MI Proxy"""

import re
import struct
import socket

from application.configuration import ConfigSection, ConfigSetting
from application.configuration.datatypes import NetworkAddress, NetworkRange, NetworkRangeList
from application.system import host
from application import log

from twisted.internet import reactor
from twisted.web2.xmlrpc import XMLRPC, NoSuchFunction, Fault
from twisted.web2.server import Site
from twisted.web2.channel import HTTPFactory
from twisted.web2.resource import Resource, LeafResource
from twisted.web2.http import StatusResponse
from twisted.python import failure

from miproxy.opensips import ManagementInterface
from miproxy import configuration_filename

class MIProxyAddress(NetworkAddress):
    default_port = 8080
    name = 'MI Proxy Address'

class URL(object):
    def __init__(self, url):
        match = re.search(r'(?P<scheme>\w+)://(?P<host>.*)', url)
        if match is None:
            raise ValueError("Illegal hostname")
        self.__dict__.update(match.groupdict())
        self.host, self.port = MIProxyAddress(self.host)

class ReplyOption(str):
    __allowed = ['string', 'array']
    def __new__(self, string):
        if string in self.__allowed:
            return string
        else:
            raise ValueError, "Illegal value for reply option"


class MIProxyConfig(ConfigSection):
    __cfgfile__ = configuration_filename
    __section__ = 'MIProxy'
    listen_url = URL("http://%s" % host.default_ip)
    xmlrpc_root = 'RPC2'
    trusted = ConfigSetting(type=NetworkRangeList, value=NetworkRangeList('any'))
    reply_option = ReplyOption('string')

if MIProxyConfig.trusted is None:
    MIProxyConfig.trusted = [NetworkRange('none')]


class RPCRoot(XMLRPC, LeafResource):
    locateChild = LeafResource.locateChild
    
    MIERROR = 8042

    def _MI_make_string(self, result):
        return result

    def _MI_make_array(self, result):
        return ['%s\n' % res for res in result.split('\n') if res]

    def _catch_all(self, command, request, *args):
        args = [str(arg) for arg in args]
        log.info("Got XMLRPC request from %s: %s (%s)" % (request.remoteAddr.host, command, ' '.join(args)))
        
        deferred = ManagementInterface().send_command(command, *args)
        deferred.addCallback(getattr(self, '_MI_make_%s' % MIProxyConfig.reply_option))
        deferred.addErrback(lambda fail: failure.Failure(Fault(self.MIERROR, str(fail.value))))
        return deferred

    def getFunction(self, functionPath):
        try:
            return XMLRPC.getFunction(self, functionPath)
        except NoSuchFunction:
            return lambda *x: self._catch_all(functionPath, *x)

    def _cbRender(self, result, request):
        if isinstance(result, Fault) and result.faultCode == self.MIERROR:
            code, value = result.faultString.split(' ', 1)
            code = int(code)
            return StatusResponse(code, value, value)
        return XMLRPC._cbRender(self, result, request)

class MIRoot(Resource):
    def __init__(self):
        self.putChild(MIProxyConfig.xmlrpc_root, RPCRoot())

    def locateChild(self, request, segements):
        child = Resource.locateChild(self, request, segements)
        for range in MIProxyConfig.trusted:
            if struct.unpack('!L', socket.inet_aton(request.remoteAddr.host))[0] & range[1] == range[0]:
                if child[0] is None:
                    log.error("Got XMLRPC request that does not match any Resource: %s" % request.uri)
                return child
        log.error("Got XMLRPC request from %s which is not a trusted host" % request.remoteAddr.host)
        return (None, [])

class MIProxy(object):
    def __init__(self):
        self.site = Site(MIRoot())

    def run(self):
        self.on_startup()
        reactor.run()
        self.on_shutdown()

    def on_startup(self):
        reactor.listenTCP(MIProxyConfig.listen_url.port, HTTPFactory(self.site), interface=MIProxyConfig.listen_url.host)

    def on_shutdown(self):
        pass
